
-- Which diffrent types of actions are available?
-- -> The following fields are part of an action:
-- 	a_id		the ID/key of the action
-- 	a_type		selected from values_what
-- 	a_value		used to store the subtype of a_type
--
-- npc_gives and npc_wants:
--	a_on_failure	if the action fails, go to this dialog
--	a_value		itemstack of the given/wanted item in string form
--	a_item_desc	the description the NPC shall set for that itemstack
--			(so that the player can distinguish it from other
--			itemstacks with the same items)
--	a_item_quest_id	Special ID to make sure that it is really the *right*
--			item and not just something the player faked with an
--			engraving table or something similar
--
-- Note: Trades are not stored as actions - they are stored in
-- dialog.trades[ trade_id ] with <trade_id> == "<d_id>  <o_id>"
--

-- some helper lists for creating the formspecs and evaulating
-- the player's answers:

-- general direction of what could make up an action
local check_what = {
	"- please select -",
	"No action (default).", -- 2
	"Normal trade - one item(stack) for another item(stack).", -- 3
	"The NPC gives something to the player (i.e. a quest item).", -- 4
	"The player is expected to give something to the NPC (i.e. a quest item).", -- 5
	"The player has to manually enter a password or passphrase or some other text.", -- 6
	"Call custom functions that are supposed to be overridden by the server.", -- 7
	-- "The player has to move virtual items in a virtual inventory to the right position.", -- 8
}

-- how to store these as a_type in the action:
local values_what = {"", "none", "trade", "npc_gives", "npc_wants", "text_input", "custom", "puzzle"}


-- monitor changes to the npc_gives and npc_wants slots (in particular when editing actions)
-- how: can be "put" or "take"
yl_speak_up.action_inv_changed = function(inv, listname, index, stack, player, how)
	if(not(player)) then
		return
	end
	local pname = player:get_player_name()
	local n_id = yl_speak_up.speak_to[pname].n_id
	-- if not in edit mode: the player may just be normally interacting with the NPC;
	-- nothing to do for us here (wait for the player to click on "save")
	if(not(n_id) or yl_speak_up.edit_mode[pname] ~= n_id) then
		return
	end
	-- is the player in the process of editing an action of the npc_gives/npc_wants type?
	local data = yl_speak_up.speak_to[pname][ "tmp_action" ]
	if(not(data) or (data.what ~= 4 and data.what ~= 5)) then
		return
	end
	-- "The NPC gives something to the player (i.e. a quest item).", -- 4
	if(    how == "put"  and data.what == 4) then
		data.item_node_name = stack:get_name().." "..stack:get_count()
		local meta = stack:get_meta()
		if(meta and meta:get_string("description")) then
			data.item_desc = meta:get_string("description")
		end
		if(meta and meta:get_string("yl_speak_up:quest_id")) then
			data.item_quest_id = meta:get_string("yl_speak_up:quest_id")
		end
	elseif(how == "take" and data.what == 4) then
		data.item_desc = "- no item set -"
		data.item_node_name = ""
	-- "The player is expected to give something to the NPC (i.e. a quest item).", -- 5
	elseif(how == "put"  and data.what == 5) then
		data.item_node_name = stack:get_name().." "..stack:get_count()
		local meta = stack:get_meta()
		if(meta and meta:get_string("description")) then
			data.item_desc = meta:get_string("description")
		end
		if(meta and meta:get_string("yl_speak_up:quest_id")) then
			data.item_quest_id = meta:get_string("yl_speak_up:quest_id")
		end
	elseif(how == "take" and data.what == 5) then
		data.item_desc = "- no item set -"
		data.item_node_name = ""
	end
	-- show the updated formspec to the player
	yl_speak_up.show_fs(player, "edit_actions", nil)
	-- no need to check anything more here; the real checks need to be done
	-- when the player presses the save/store/execute button
end


-- returns a human-readable text as description of the action
-- (as shown in the edit options dialog and in the edit effect formspec)
yl_speak_up.show_action = function(a)
	if(not(a.a_type) or a.a_type == "" or a.a_type == "none") then
		return "(nothing): Nothing to do. No action."
	elseif(a.a_type == "trade") then
		return "NPC sells \""..table.concat(a.a_buy, ";").."\" for \""..
			table.concat(a.a_pay, ";").."\"."
	elseif(a.a_type == "npc_gives") then
		return "The NPC gives \""..tostring(a.a_item_desc or "- default description -")..
			"\" (\""..tostring(a.a_value or "- ? -").."\") "..
			"with ID \""..tostring(a.a_item_quest_id or "- no special ID -").."\"."
	elseif(a.a_type == "npc_wants") then
		return "The NPC wants \""..tostring(a.a_item_desc or "- default description -")..
			"\" (\""..tostring(a.a_value or "- ? -").."\") "..
			"with ID \""..tostring(a.a_item_quest_id or "- no special ID -").."\"."
	elseif(a.a_type == "text_input") then
		return "Q: \""..tostring(a.a_question).."\" A:\""..tostring(a.a_value).."\"."
-- puzzle is unused; better do that via custom
--	elseif(a.a_type == "puzzle") then
--		return "puzzle:"
	elseif(a.a_type == "custom") then
		return "Call custom function with param: \""..tostring(a.a_value).."\"."
	end
	-- fallback
	return tostring(a.a_value)
end


-- actions - in contrast to preconditions and effects - may take time
-- because the player usually gets presented a formspec and needs to
-- react to that; thus, we can't just execute all actions simultaneously
yl_speak_up.execute_next_action = function(player, a_id, result_of_a_id)
	local pname = player:get_player_name()
	local n_id = yl_speak_up.speak_to[pname].n_id
	local d_id = yl_speak_up.speak_to[pname].d_id
	local o_id = yl_speak_up.speak_to[pname].o_id
	local dialog = yl_speak_up.speak_to[pname].dialog
	yl_speak_up.debug_msg(player, n_id, o_id, "Last action: "..tostring(a_id).." returned "..
				tostring(result_of_a_id)..".")
	local actions = {}
	local effects = {}
	local sorted_key_list = {}
	if(dialog
	  and dialog.n_dialogs
	  and dialog.n_dialogs[d_id]
	  and dialog.n_dialogs[d_id].d_options
	  and dialog.n_dialogs[d_id].d_options[o_id]
	  and dialog.n_dialogs[d_id].d_options[o_id].actions) then
		-- get the actual actions
		actions = dialog.n_dialogs[d_id].d_options[o_id].actions
		-- needed later on when all actions are executed
		effects = dialog.n_dialogs[d_id].d_options[o_id].o_results
		-- sort the actions so that we can execute them always in the
		-- same order
		sorted_key_list = yl_speak_up.sort_keys(actions)
		local nr = 0
		if(a_id) then
			nr = table.indexof(sorted_key_list, a_id)
			-- did the action fail?
			if(nr > -1 and not(result_of_a_id)) then
				local this_action = actions[ sorted_key_list[ nr ]]
				-- if there is an on_failure target dialog: go there
				if(this_action.a_on_failure) then
					-- set the current action to nil
					yl_speak_up.speak_to[pname].a_id = nil
					-- no option of the new dialog has been selected yet
					yl_speak_up.speak_to[pname].o_id = nil
					-- show the new dialog
					yl_speak_up.debug_msg(player, n_id, o_id, "Action "..
						tostring(a_id).." failed. Switching to dialog "..
						tostring(this_action.a_on_failure)..".")
					yl_speak_up.show_fs(player, "talk", {n_id = n_id,
						d_id = this_action.a_on_failure})
					return
				else
					yl_speak_up.debug_msg(player, n_id, o_id, "Action "..
						tostring(a_id).." failed, but no a_on_failure target "..
						"dialog defined. Continuing execution.")
				end
			end
		end
		-- get the next entry
		if(nr > -1 and nr < #sorted_key_list and sorted_key_list[nr + 1]) then
			local next_action = actions[ sorted_key_list[ nr + 1 ]]
			-- store which action we are currently executing
			yl_speak_up.speak_to[pname].a_id = next_action.a_id
			-- execute the next action
			yl_speak_up.debug_msg(player, n_id, o_id, "Executing next action "..
				tostring(next_action.a_id)..".")
			yl_speak_up.execute_action(player, n_id, o_id, next_action)
			-- the player needs time to react
			return
		end
	end
	-- when all actions are executed:
	-- set the current action to nil
	yl_speak_up.speak_to[pname].a_id = nil
	yl_speak_up.debug_msg(player, n_id, o_id, "All actions have been executed successfully. "..
		"Doing effects/results now.")
	-- execute all effects/results
	local target_dialog = yl_speak_up.execute_all_relevant_effects(player, effects, o_id, true)
	-- the function above returns a target dialog; show that to the player
	yl_speak_up.show_fs(player, "talk", {n_id = n_id, d_id = target_dialog})
end


yl_speak_up.execute_action = function(player, n_id, o_id, a)
	if(not(a.a_type) or a.a_type == "" or a.a_type == "none") then
		-- no action - nothing to do
		return true
	elseif(a.a_type == "trade") then
		yl_speak_up.show_fs(player, "trade_simple", a.a_value)
		return
	elseif(a.a_type == "npc_gives") then
		-- TODO implement npc_gives action
	elseif(a.a_type == "npc_wants") then
		-- TODO implement npc_wants action
	elseif(a.a_type == "text_input") then
		-- TODO implement text_input action
	elseif(a.a_type == "custom") then
		-- TODO implement custom action
	end
	-- fallback: unkown type
	return false
end


-- these are only wrapper functions for those in fs_edit_general.lua

yl_speak_up.input_fs_edit_actions = function(player, formname, fields)
	return yl_speak_up.input_fs_edit_option_related(player, formname, fields,
		"a_", "actions", yl_speak_up.max_actions,
		"(A)ctions", "tmp_action",
		nil, -- unused - no block operations
		values_what, {}, {}, {}, {},
		check_what, {}, {}, {}, {},
		nil, -- no variables
		"edit_actions"
		)
end

yl_speak_up.get_fs_edit_actions = function(player, table_click_result)
	return yl_speak_up.get_fs_edit_option_related(player, table_click_result,
		"a_", "actions", yl_speak_up.max_actions,
		"(A)ctions", "tmp_action",
		"What do you want to happen in this (A)ction?",
		check_what, {}, {}, {}, {},
		nil, -- no variables
		yl_speak_up.show_action,
		"table_of_elements",
		nil, nil, nil, -- no variable handling here
		nil -- nothing block-related to do here
		)
end
